import requests
from interfaces import *

class APIService():
    def __init__(self, url):
        self.server_url = url
        self.auth_token = None
        return
    
    def get(self, service_url, query, auth_required=False):
        """
            Makes a GET request and returns the results
            Inputs:
                - service_url(str): the url to make the request to.
                - query(dict): the get query.
                - auth_required(bool): indicates whether authorization is required for the call or not.  
        """
        url = self._create_url(service_url)
        response = None
        if auth_required:
            response = requests.get(url=url, params=query, headers={'Authorization': 'Bearer {}'.format(self.auth_token)})
        else:
            response = requests.get(url, params=query)

        try:
            return response.json()
        except:
            raise Exception("Error while getting {}: {}".format(url, response))
        
    def post(self, service_url, data, auth_required=False):
        """
            Makes a POST request and returns the results
            Inputs:
                - service_url(str): the url to make the request to.
                - data(dict): the data to post with POST request.
                - auth_required(bool): indicates whether authorization is required for the call or not.
        """
        url = self._create_url(service_url)
        response = None
        if auth_required:
            response = requests.post(url=url, data=data, headers={'Authorization': 'Bearer {}'.format(self.auth_token)})

        else:
            response = requests.post(url=url, data=data)

        if "error" in response:
            raise Exception("Error while posting to {}: {}".format(url, response))
        else:
            return response.json()

    def delete(self,service_url, data, auth_required=False):
        """
            Makes a DELETE request and returns the results
            Inputs:
                - service_url(str): the url to make the request to.
                - data(dict): the data to post with POST request.
                - auth_required(bool): indicates whether authorization is required for the call or not.
        """
        url = self._create_url(service_url)
        if auth_required:
            return requests.delete(url=url, data=data, headers={'Authorization': 'Bearer {}'.format(self.auth_token)}).json()
        else:
            return requests.delete(url=url, data=data).json()
 
    '''
        Private methods
    '''
    def _create_url(self, path):
        """
            Appends namespace to server url
        """
        return "{}{}".format(self.server_url, path)



import socketio
from enumerations import MARKET_SYMBOLS, SOCKET_EVENTS

sio = socketio.Client()
      
class Sockets:
    callbacks={}
    def __init__(self, url, timeout=10, token=None) -> None:
        self.url = url  
        self.timeout = timeout
        self.token = token
        return 

    def _establish_connection(self):
        """
            Connects to the desired url
        """
        try:
            sio.connect(self.url,wait_timeout=self.timeout,transports=["websocket"])
            return True
        except:
            return False

    def set_token(self, token):
        """
            Sets default user token
            Inputs:
                - token (user auth token): firefly onboarding token.
        """
        self.token = token

    async def open(self):
        """
            opens socket instance connection
        """
        self.connection_established = self._establish_connection()
        if not self.connection_established:
            self.close()
            raise(Exception("Failed to connect to Host: {}".format(self.url)))
        return
        

    async def close(self):
        """
            closes the socket instance connection
        """
        sio.disconnect()
        return 

    @sio.on("*")
    def listener(event,data):
        """
            Listens to all events emitted by the server
        """
        try:
            if event in Sockets.callbacks.keys():
                Sockets.callbacks[event](data)
            elif "default" in Sockets.callbacks.keys():
                Sockets.callbacks["default"]({"event":event,"data":data})
            else:
                pass
        except:
            pass
        return 

    async def listen(self,event,callback):
        """
            Assigns callbacks to desired events
        """
        Sockets.callbacks[event] = callback
        return 

    async def subscribe_global_updates_by_symbol(self,symbol: MARKET_SYMBOLS):
        """
            Allows user to subscribe to global updates for the desired symbol.
            Inputs:
                - symbol: market symbol of market user wants global updates for. (e.g. DOT-PERP)
        """
        try:
            if not self.connection_established:
                raise Exception("Socket connection is established, invoke socket.open()")

            sio.emit('SUBSCRIBE',[
            {
                "e": SOCKET_EVENTS.GLOBAL_UPDATES_ROOM.value,
                "p": symbol.value,
            },
            ])
            return True
        except Exception as e:
            print("Error: ", e)
            return False

    async def unsubscribe_global_updates_by_symbol(self,symbol: MARKET_SYMBOLS):
        """
            Allows user to unsubscribe to global updates for the desired symbol.
                Inputs:
                    - symbol: market symbol of market user wants to remove global updates for. (e.g. DOT-PERP)
        """
        try:
            if not self.connection_established:
                return False 
            
            sio.emit('UNSUBSCRIBE', [
            {
                "e": SOCKET_EVENTS.GLOBAL_UPDATES_ROOM.value,
                "p": symbol.value,
            },
            ])
            return True
        except:
            return False

    async def subscribe_user_update_by_token(self,user_token: str=None):
        """
            Allows user to subscribe to their account updates.
            Inputs:
                - token: auth token generated when onboarding on firefly
        """
        try:
            if not self.connection_established:
                return False
              
            sio.emit("SUBSCRIBE", [
            {
                "e": SOCKET_EVENTS.USER_UPDATES_ROOM.value,
                "t": self.token if user_token == None else user_token,
            },
            ])
            return True
        except:
            return False

    async def unsubscribe_user_update_by_token(self,user_token:str=None): 
        """
            Allows user to unsubscribe to their account updates.
            Inputs:
                - token: auth token generated when onboarding on firefly
        """
        try:
            if not self.connection_established:
                return False
              
            sio.emit("UNSUBSCRIBE", [
            {
                "e": SOCKET_EVENTS.USER_UPDATES_ROOM.value,
                "t": self.token if user_token == None else user_token,
            },
            ])
            return True
        except:
            return False

    
  


from typing import TypedDict
from enumerations import *

class Order(TypedDict):
    isBuy: bool 
    reduceOnly: bool 
    quantity: int 
    price: int 
    triggerPrice: int 
    leverage: int 
    expiration: int 
    salt: int   
    maker: str 

class SignedOrder(Order):
    typedSignature: str

class RequiredOrderFields(TypedDict): 
  symbol: MARKET_SYMBOLS # market for which to create order
  price: int # price at which to place order. Will be zero for a market order
  quantity: int # quantity/size of order
  side: ORDER_SIDE # BUY/SELL
  orderType: ORDER_TYPE # MARKET/LIMIT


class OrderSignatureRequest(RequiredOrderFields): 
  leverage: int # (optional) leverage to take, default is 1
  reduceOnly: bool # (optional)  is order to be reduce only true/false, default its false
  salt: int # (optional)  random number for uniqueness of order. Generated randomly if not provided
  expiration: int # (optional) time at which order will expire. Will be set to 1 month if not provided
  maker: str # (optional) maker of the order, if not provided the account used to initialize the client will be default maker
  
class OrderSignatureResponse(RequiredOrderFields):
  maker: str 
  orderSignature: str

class PlaceOrderRequest(OrderSignatureResponse):
  timeInForce: TIME_IN_FORCE # FOK/IOC/GTT by default all orders are GTT
  postOnly: bool # true/false, default is true
  clientId: str # id of the client

class GetOrderbookRequest(TypedDict):
  symbol: str
  limit: int # number of bids/asks to retrieve, should be <= 50

class OnboardingMessage(TypedDict):
    action: str
    onlySignOn: str

class OrderResponse(TypedDict):
  id: int
  clientId: str
  requestTime: int
  cancelReason: CANCEL_REASON
  orderStatus: ORDER_STATUS
  hash: str
  symbol: MARKET_SYMBOLS
  orderType: ORDER_TYPE
  timeInForce: TIME_IN_FORCE
  userAddress: str
  side: ORDER_SIDE
  price: str
  quantity: str
  leverage: str
  reduceOnly: bool
  expiration: int
  salt: int
  orderSignature: str
  filledQty: str
  avgFillPrice: str
  createdAt: int
  updatedAt: int
  makerFee: str
  takerFee: str
  openQty: str
  cancelOnRevert: bool


class GetOrderResponse(OrderResponse):
  fee: str
  postOnly: bool
  triggerPrice: str


class GetCandleStickRequest(TypedDict):
  symbol: MARKET_SYMBOLS
  interval: Interval
  startTime: float
  endTime: float
  limit: int

class GetMarketRecentTradesRequest(TypedDict):
  symbol: MARKET_SYMBOLS
  pageSize: int
  pageNumber: int
  traders: TRADE_TYPE

class OrderCancelSignatureRequest(TypedDict):
  symbol: MARKET_SYMBOLS
  hashes: list

class OrderCancellationRequest(OrderCancelSignatureRequest):
  signature: str

class CancelOrder(TypedDict):
  hash: str
  reason: str


class CancelOrderResponse(TypedDict):
  message: str
  data: dict


class GetTransactionHistoryRequest(TypedDict):
  symbol: MARKET_SYMBOLS  # will fetch orders of provided market
  pageSize: int  # will get only provided number of orders must be <= 50
  pageNumber: int  # will fetch particular page records. A single page contains 50 records.

class GetPositionRequest(TypedDict):
  symbol: MARKET_SYMBOLS  # will fetch orders of provided market
  pageSize: int  # will get only provided number of orders must be <= 50
  pageNumber: int  # will fetch particular page records. A single page contains 50 records.

class GetUserTradesRequest(TypedDict):
  symbol: MARKET_SYMBOLS
  maker: bool
  fromId: int
  startTime: int
  endTime: int
  pageSize: int
  pageNumber: int
  type: ORDER_TYPE

class GetOrderRequest(GetTransactionHistoryRequest):
  statuses:ORDER_STATUS # status of orders to be fetched





from datetime import datetime
from random import randint
from web3 import Web3


def strip_hex_prefix(input):
    if input[0:2] == '0x':
        return input[2:]
    else:
        return input

def address_to_bytes32(addr):
    return '0x000000000000000000000000' + strip_hex_prefix(addr)


def hash_string(value: str): 
    return Web3.soliditySha3(["string"], [value] ).hex()

def bn_to_bytes8(value:int):
    return str("0x"+"0"*16+hex(value)[2:]).encode('utf-8')

def default_value(dict, key, default_value):
    if key in dict:
        return dict[key]
    else:
        return default_value 

def default_enum_value(dict, key, default_value):
    if key in dict:
        return dict[key].value
    else:
        return default_value.value


def current_unix_timestamp():
        return int(datetime.now().timestamp())

def random_number(max_range):
    return current_unix_timestamp() + randint(0, max_range) + randint(0, max_range)

def extract_query(value:dict):
    query=""
    for i,j in value.items():
        query+="&{}={}".format(i,j)
    return query[1:]

def extract_enums(params:dict,enums:list):
    for i in enums:
        if i in params.keys():
            if type(params[i]) == list:
                params[i] = [x.value for x in params[i]]
            else:
                params[i] = params[i].value
    return params